# Copyright (c) 2015, Sergiusz Bazanski <q3k@q3k.org>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from functools import wraps

import memcache
import requests
import sqltap.wsgi

from flask import Flask, redirect
from flask_sqlalchemy import SQLAlchemy
from flask_login import LoginManager, AnonymousUserMixin, login_required, current_user
from flask_caching import Cache
from flaskext.gravatar import Gravatar

app = Flask(__name__)
app.config.from_object("config.CurrentConfig")

db = SQLAlchemy()
login_manager = LoginManager()
cache = Cache()
gravatar = Gravatar(size=256, rating='g', default='retro',
                    force_default=False, use_ssl=True, base_url=None)

# TODO unsubscribe me from life
cache_enabled = False
mc = cache


import webapp.models  # noqa
from webapp.auth import AnonymousUser, User

login_manager.anonymous_user = AnonymousUser

@login_manager.user_loader
def load_user(username):
    return User(username)

@login_manager.unauthorized_handler
def unauthorized():
    return redirect('/login')

def admin_required(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if not current_user.is_admin():
            return login_manager.unauthorized()
        return func(*args, **kwargs)
    return wrapper


import webapp.views
import webapp.admin
import webapp.api



def init():
    db.init_app(app)
    login_manager.init_app(app)
    gravatar.init_app(app)
    cache.init_app(app)

    # Initialize middleware
    app.wsgi_app = sqltap.wsgi.SQLTapMiddleware(app.wsgi_app)

    # Register blueprints
    app.register_blueprint(webapp.admin.bp)
    app.register_blueprint(webapp.api.bp)

    # Custom filters
    @app.template_filter('inflect')
    def inflect(v, one, two, five):
        num = abs(v)

        if num == 0:
            return '%d %s' % (v, five)
        elif num == 1:
            return '%d %s' % (v, one)
        elif num <= 4:
            return '%d %s' % (v, two)

        return '%d %s' % (v, five)
