# - * - coding=utf-8 - * -

import datetime
from email.mime.text import MIMEText
from subprocess import Popen, PIPE

from flask import render_template, request, flash, g, Response, \
    redirect, url_for, abort, Blueprint
from flask_login import login_required
from webapp import forms, db, models, admin_required

import directory
import logic


bp = Blueprint('admin', __name__)

@bp.route("/admin")
@admin_required
@login_required
def index():
    members = [m.get_status() for m in models.Member.get_members(True)]
    for member in members:
        due = member['months_due']
        if due < 1:
            member['color'] = "00FF00"
        elif due < 3:
            member['color'] = "E0941B"
        else:
            member['color'] = "FF0000"

    active_members = filter(lambda m: m['judgement'], members)
    inactive_members = filter(lambda m: not m['judgement'], members)
    diff = directory.get_ldap_group_diff(members)
    if diff is not None:
        flash("LDAP sync required")
    return render_template("admin_index.html",
                           active_members=active_members,
                           inactive_members=inactive_members)


@bp.route("/admin/ldapsync", methods=["POST", "GET"])
@admin_required
@login_required
def admin_ldap_sync():
    members = [m.get_status() for m in models.Member.get_members(True)]
    diff = directory.get_ldap_group_diff(members)
    if diff is None:
        return render_template("admin_ldap_sync.html", form=False)

    form = forms.LDAPSyncForm(request.form)

    form.fatty_to_add.choices = zip(diff['fatty_to_add'], diff['fatty_to_add'])
    form.fatty_to_add.default = diff['fatty_to_add']

    form.fatty_to_remove.choices = zip(diff['fatty_to_remove'], diff['fatty_to_remove'])
    form.fatty_to_remove.default = diff['fatty_to_remove']

    form.starving_to_add.choices = zip(diff['starving_to_add'], diff['starving_to_add'])
    form.starving_to_add.default = diff['starving_to_add']

    form.starving_to_remove.choices = zip(diff['starving_to_remove'], diff['starving_to_remove'])
    form.starving_to_remove.default = diff['starving_to_remove']

    form.process(request.form)
    if form.validate():
        changes = {'fatty': {}, 'starving': {}}
        changes['fatty']['add'] = form.fatty_to_add.data
        changes['fatty']['remove'] = form.fatty_to_remove.data
        changes['starving']['add'] = form.starving_to_add.data
        changes['starving']['remove'] = form.starving_to_remove.data

        directory.update_member_groups(g.ldap, changes)

    return render_template("admin_ldap_sync.html", form=form)


@bp.route("/admin/csv")
@admin_required
@login_required
def admin_csv():
    members = []
    for m in models.Member.get_members(True):
        member = m.get_status()
        if member['type'] == 'supporting':
            continue
        member['contact_email'] = m.get_contact_email()
        member['cn'] = directory.get_member_fields(g.ldap, member['username'], 'cn')['cn']
        members.append(member)

    active_members = filter(lambda m: m['judgement'], members)
    output = render_template("admin_csv.html", active_members=active_members)
    return Response(output)

@bp.route('/admin/member/<membername>')
@login_required
@admin_required
def admin_member(membername):
    member = models.Member.get_members(True).filter_by(username=membername).first()
    if not member:
        abort(404)
    status = member.get_status()
    cn = directory.get_member_fields(g.ldap, member.username, 'cn')['cn']
    return render_template("admin_member.html", member=member, status=status,
                           cn=cn, admin=True)

@bp.route("/admin/member/<membername>/policy:<policy>")
@login_required
@admin_required
def admin_member_set_policy(membername, policy):
    member = models.Member.query.filter_by(username=membername).first()
    member.payment_policy = models.PaymentPolicy[policy].value
    db.session.add(member)
    db.session.commit()
    return redirect(request.referrer)

@bp.route("/admin/member/<membername>/membership:<membershiptype>")
@login_required
@admin_required
def admin_member_set_membership(membername, membershiptype):
    member = models.Member.query.filter_by(username=membername).first()
    member.type = models.MembershipType[membershiptype].name
    db.session.add(member)
    db.session.commit()
    return redirect(request.referrer)


@bp.route("/admin/member/add/<membershiptype>/<username>")
@login_required
@admin_required
def add_member(membershiptype, username):
    member = models.Member(None, username, models.MembershipType[membershiptype].name, True)
    db.session.add(member)
    db.session.commit()
    flash('Member created')
    return redirect(request.referrer)

@bp.route("/admin/match")
@login_required
@admin_required
def match_index():
    transfers_unmatched = logic.get_unmatched_transfers()
    return render_template("match.html", transfers_unmatched=transfers_unmatched)


@bp.route("/admin/match/auto", methods=["GET"])
@login_required
@admin_required
def match_auto():
    matched = 0
    left = 0
    transfers_unmatched = logic.get_unmatched_transfers()
    affected_members = []
    for transfer in transfers_unmatched:
        matchability, member, months = transfer.get_matchability()
        try:
            print "[i] Matching transfer {} for {:.2f}PLN by member {}, {} months".format(
                transfer.id, transfer.amount/100, member.username, months)
        except AttributeError:
            print "[e] Member data invalid, WTF - {}".format(repr(member))
            continue
        if matchability == models.Transfer.MATCH_OK:
            if member.transfers:
                year, month = member.get_next_unpaid()
                if None in (year, month):
                    print "[w] next_unpaid borked, skipping"
                    continue
            else:
                year, month = transfer.date.year, transfer.date.month
            for m in range(months):
                mt = models.MemberTransfer(None, year, month, transfer)
                member.transfers.append(mt)
                db.session.add(mt)
                flash("Matched transfer {} for {:.2f}PLN to member {} for month {}-{}".format(transfer.id, transfer.amount/100, member.username, year, month))
                year, month = member._yearmonth_increment((year,month))
            matched += 1
            affected_members.append(member)
        else:
            left += 1
    db.session.commit()
    for member in affected_members:
        member.get_status(force_refresh=True)
    flash("Matched %i, %i left" % (matched, left))
    return redirect(url_for(".match_index"))

@bp.route("/admin/match/manual", methods=["GET"])
@login_required
@admin_required
def match_manual():
    transfers_unmatched = logic.get_unmatched_transfers()
    return render_template("match_manual.html", transfers_unmatched=transfers_unmatched)

@bp.route("/admin/match/ignore/<path:uid>")
@login_required
@admin_required
def ignore(uid):
    transfer = models.Transfer.query.filter_by(uid=uid).first()
    if not transfer:
        flash('No transfer found', 'danger')
        return redirect(url_for(".match_manual"))
    transfer.ignore = True
    db.session.commit()

    flash('Transfer %s ignored' % (transfer,))

    return redirect(request.referrer)

@bp.route("/admin/match/<username>/<int:months>/<path:uid>")
@login_required
@admin_required
def match(username, uid, months):
    member = models.Member.query.filter_by(username=username).first()
    if not member:
        flash('No member found', 'danger')
        return redirect(url_for(".match_manual"))

    transfer = models.Transfer.query.filter_by(uid=uid).first()
    if not transfer:
        flash('No transfer found', 'danger')
        return redirect(url_for(".match_manual"))

    for _ in range(months):
        year, month = member.get_next_unpaid()
        mt = models.MemberTransfer(None, year, month, transfer)
        member.transfers.append(mt)
        db.session.add(mt)

    db.session.commit()
    member.get_status(force_refresh=True)

    flash('OK, %d get' % transfer.amount)
    return redirect(url_for(".match_manual"))


@bp.route("/admin/match/", methods=["POST"])
@login_required
@admin_required
def match_user_transfer():
    username = request.form["username"]
    uid = request.form["uid"]
    member = models.Member.query.filter_by(username=username).first()
    if not member:
        flash('No member found', 'danger')
        return redirect(url_for(".match_manual"))

    transfer = models.Transfer.query.filter_by(uid=uid).first()
    if not transfer:
        flash('No transfer found', 'danger')
        return redirect(url_for(".match_manual"))

    return render_template("match_user_transfer.html", member=member, transfer=transfer)

@bp.route("/admin/spam/", methods=["GET", "POST"])
@login_required
@admin_required
def sendspam():
    now = datetime.datetime.now()
    members = models.Member.query.filter_by(
        active=True, payment_policy=models.PaymentPolicy.normal.value).all()

    form = forms.SpamForm()
    form.members.choices = [(member.id, member) for member in members]
    form.members.default = [member.id for member in members]

    form.process(request.form)

    if form.validate():
        spam = []
        for member in members:
            if member.id not in form.members.data:
                continue

            content = render_template(
                'mailing/due.txt',
                member=member,
                status=member.get_status(),
                transfers=member.transfers[:5],
                now=now)

            # Just ignore empty messages
            if not content.strip():
                continue

            msg = MIMEText(content, "plain", "utf-8")
            msg["From"] = "Kasownik Hackerspace'owy <kasownik@hackerspace.pl>"
            msg["Subject"] = "Stan składek na dzień %s" % now.strftime("%d/%m/%Y")
            msg["To"] = member.get_contact_email()
            spam.append(msg)

        if form.dry_run.data:
            readable = [
                msg.as_string().split('\n\n')[0] + '\n\n'
                + msg.get_payload(decode=True) for msg in spam]
            return Response('\n====\n'.join(readable), mimetype='text/text')

        for msg in spam:
            p = Popen(["/usr/sbin/sendmail", "-t"], stdin=PIPE)
            p.communicate(msg.as_string())

        flash('%d messages sent!' % len(spam))
        return redirect(url_for('.index'))
    return render_template('admin_spam.html', form=form)
